/*
 * Copyright © 2013-2014 Alexander Söderlund, Sweden
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>		/* assert */
#include <limits.h>		/* *_MAX */
#include <string.h>		/* strcmp */

#include "gleipnir_line.h"	/* gleip_line_* */
#include "gleipnir_lines.h"	/* gleip_lines_* */
#include "gleipnir_lineread.h"	/* gleip_lineread_* */

#include "config.h"		/* HNEF_RSTR */
#include "func.h"		/* hnef_fr_good */
#include "game.h"		/* hnef_game_*, ... */
#include "movet.h"		/* HNEF_BOARDPOS_NONE */
#include "num.h"		/* hnef_texttoushort */
#include "rread.h"
#include "rreadt.h"		/* HNEF_RREAD_L_* */
#include "ruleset.h"		/* hnef_ruleset_valid_size */
#include "rvalid.h"		/* hnef_game_valid */
#include "type_piece.h"		/* hnef_type_piece_* */
#include "type_square.h"	/* hnef_type_square_* */
#include "types.h"		/* HNEF_TYPEC_* */

/*
 * Reads "width" or "height" into the "size" pointer.
 *
 * "size" is a pointer to rules->bwidth or rules->bheight.
 */
static
enum HNEF_FR
hnef_parse_size (
/*@in@*/
/*@notnull@*/
	const struct hnef_ruleset	* const HNEF_RSTR rules,
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@in@*/
/*@notnull@*/
	unsigned short	* const HNEF_RSTR size,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD	* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * size, * fail_read@*/
{
	assert	(NULL != rules);
	assert	(NULL != wordv);
	assert	(NULL != size);
	assert	(NULL != fail_read);

	if	(NULL != rules->pieces
	||	NULL != rules->squares)
	{
		* fail_read	= HNEF_RREAD_FAIL_ARRB_PRESIZE;
		return		HNEF_FR_SUCCESS;
	}

	if	(wordc < (size_t)2
/*@i2@*/\
	||	gleip_line_empty(wordv[1]))
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}
/*@i2@*/\
	else if	(HNEF_BOARDPOS_NONE != * size)
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_DUP;
		return		HNEF_FR_SUCCESS;
	}

	if	(!hnef_texttoushort(wordv[1], size))
	{
		* size		= HNEF_BOARDPOS_NONE;
		* fail_read	= HNEF_RREAD_FAIL_STRTOL;
		return		HNEF_FR_SUCCESS;
	}
/*@i2@*/\
	else if	(!hnef_ruleset_valid_size(* size))
	{
		* size		= HNEF_BOARDPOS_NONE;
		* fail_read	= HNEF_RREAD_FAIL_NOOB;
		return		HNEF_FR_SUCCESS;
	}
	return	HNEF_FR_SUCCESS;
}

/*
 * Reads the pieces or squares array in ruleset. If read_sq, then read
 * game->rules->squares; else read game->rules->pieces.
 *
 * This requires that width and height are set. If they are not (i.e. if
 * they are invalid), then it fails.
 */
static
enum HNEF_FR
hnef_parse_arrb (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@partial@*/
/*@notnull@*/
	struct hnef_game	* const HNEF_RSTR game,
	const HNEF_BOOL		read_sq,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * game, * fail_read@*/
{
	struct hnef_ruleset	* HNEF_RSTR rules	= NULL;
	unsigned short		num,
				blen;
	size_t			i;
	HNEF_BIT_U8 * HNEF_RSTR	arr;

	assert	(NULL != wordv);
	assert	(NULL != game);
	assert	(NULL != fail_read);

	rules	= game->rules;

	if	(!hnef_ruleset_valid_size(rules->bwidth)
	||	!hnef_ruleset_valid_size(rules->bheight))
	{
		* fail_read	= HNEF_RREAD_FAIL_NOPREL_BSIZE;
		return		HNEF_FR_SUCCESS;
	}
	blen	= (unsigned short)(rules->bwidth * rules->bheight);

	if	(wordc - 1 != (size_t)blen)
	{
		/*
		 * Too many or too few arguments.
		 */
		* fail_read	= HNEF_RREAD_FAIL_ARRB_SIZE;
		return		HNEF_FR_SUCCESS;
	}

	if	(read_sq)
	{
		if	(NULL != rules->squares)
		{
			/*
			 * Already read (or not cleared from before).
			 */
			* fail_read	= HNEF_RREAD_FAIL_ARG_DUP;
			return		HNEF_FR_SUCCESS;
		}
		rules->squares	= malloc(sizeof(* rules->squares)
				* (size_t)blen);
		arr	= rules->squares;
	}
	else
	{
		if	(NULL != rules->pieces)
		{
			/*
			 * Already read (or not cleared from before).
			 */
			* fail_read	= HNEF_RREAD_FAIL_ARG_DUP;
			return		HNEF_FR_SUCCESS;
		}
		rules->pieces	= malloc(sizeof(* rules->pieces)
				* (size_t)blen);
		arr	= rules->pieces;
	}
	if	(NULL == arr)
	{
		return	HNEF_FR_FAIL_ALLOC;
	}

	/*
	 * First word is "pieces" / "squares" argument, so skip it.
	 */
	for	(i = (size_t)1; i < wordc; ++i)
	{
		if	(!hnef_texttoushort(wordv[i], & num))
		{
			* fail_read	= HNEF_RREAD_FAIL_STRTOL;
			return		HNEF_FR_SUCCESS;
		}
		else if	(num > (unsigned short)128
		|| (0 != num && !hnef_single_bit((unsigned int)num)))
		{
			* fail_read	= HNEF_RREAD_FAIL_BIT_SINGLE;
			return		HNEF_FR_SUCCESS;
		}
		else if	(i - 1 >= (size_t)blen)
		{
			* fail_read	= HNEF_RREAD_FAIL_ARG_EXCESS;
			return		HNEF_FR_SUCCESS;
		}
		else
		{
			arr[i - 1]	= (HNEF_BIT_U8)num;
		}
	}
	return	HNEF_FR_SUCCESS;
}

/*
 * Parses "piece/square bit <arg>".
 *
 * The <arg> is added as a type_piece or type_square in game->rules.
 *
 * If read_sq, then it's a square type, else it's a piece type.
 */
static
enum HNEF_FR
hnef_parse_type_bit (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@in@*/
/*@notnull@*/
	struct hnef_game	* const HNEF_RSTR game,
	const HNEF_BOOL		read_sq,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * game, * fail_read@*/
{
	const char	* HNEF_RSTR arg2;
	unsigned short	num;

	assert	(NULL != wordv);
	assert	(NULL != game);
	assert	(NULL != fail_read);

	if	(wordc < (size_t)3
/*@i2@*/\
	||	gleip_line_empty(wordv[2]))
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}
/*@i1@*/\
	else if	(game->rules->type_piecec >= HNEF_TYPE_MAX)
	{
		* fail_read	= HNEF_RREAD_FAIL_TYPE_OVERFLOW;
		return		HNEF_FR_SUCCESS;
	}
	arg2	= wordv[2];

	if	(!hnef_texttoushort(arg2, & num))
	{
		* fail_read	= HNEF_RREAD_FAIL_STRTOL;
		return		HNEF_FR_SUCCESS;
	}
	else if	(num > (unsigned short)128
	||	!hnef_single_bit((unsigned int)num))
	{
		* fail_read	= HNEF_RREAD_FAIL_BIT_SINGLE;
		return		HNEF_FR_SUCCESS;
	}
	else
	{
		const HNEF_BIT_U8 bit	= (HNEF_BIT_U8)num;
		enum HNEF_FR fr		= HNEF_FR_SUCCESS;
		if	(read_sq)
		{
			fr = hnef_type_square_set(game->rules, bit);
		}
		else
		{
			fr = hnef_type_piece_set(game->rules, bit);
		}

		if	(HNEF_FR_FAIL_ILL_ARG == fr)
		{
			* fail_read	= HNEF_RREAD_FAIL_BIT_SINGLE;
			return		HNEF_FR_SUCCESS;
		}
		else if	(HNEF_FR_FAIL_ILL_STATE == fr)
		{
			* fail_read	= HNEF_RREAD_FAIL_TYPE_DUP;
			return		HNEF_FR_SUCCESS;
		}
		else if	(HNEF_FR_FAIL_NULLPTR == fr)
		{
			* fail_read	= HNEF_RREAD_FAIL_TYPE_ORDER;
			return		HNEF_FR_SUCCESS;
		}
		else
		{
			return	fr;
		}
	}
}

/*
 * Parses "piece/square x <arg>", where <arg> is a HNEF_BIT_U8 (pointed
 * to by bit).
 *
 * Only makes sure that the bit can be converted from a string. Does not
 * validate the value in any way.
 *
 * bit is unchanged if it fails.
 */
static
enum HNEF_FR
hnef_parse_type_bitu8 (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR argame,
/*@in@*/
/*@notnull@*/
	HNEF_BIT_U8	* const HNEF_RSTR bit,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD	* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * bit, * fail_read@*/
{
	unsigned short num;

	assert	(NULL != argame);
	assert	(NULL != bit);
	assert	(NULL != fail_read);

	if	(!hnef_texttoushort(argame, & num))
	{
		* fail_read	= HNEF_RREAD_FAIL_STRTOL;
		return		HNEF_FR_SUCCESS;
	}
/*@i1@*/\
	else if	(num > (unsigned short)UCHAR_MAX)
	{
		* fail_read	= HNEF_RREAD_FAIL_NOOB;
		return		HNEF_FR_SUCCESS;
	}
	* bit	= (HNEF_BIT_U8)num;
	return	HNEF_FR_SUCCESS;
}

/*
 * Parses "piece/square x <arg>", where <arg> is an unsigned short
 * (pointed to by shrt).
 *
 * Only makes sure that the number can be converted from a string. Does
 * not validate the value in any way.
 */
static
enum HNEF_FR
hnef_parse_type_ushrt (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR argame,
/*@in@*/
/*@notnull@*/
	unsigned short	* const HNEF_RSTR shrt,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD	* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * shrt, * fail_read@*/
{
	assert	(NULL != argame);
	assert	(NULL != shrt);
	assert	(NULL != fail_read);

	if	(!hnef_texttoushort(argame, shrt))
	{
		* fail_read	= HNEF_RREAD_FAIL_STRTOL;
		return		HNEF_FR_SUCCESS;
	}
	return	HNEF_FR_SUCCESS;
}

/*
 * Parses "piece x <arg>", where x is a single bit, and tp is the piece
 * corresponding to bit x.
 */
static
enum HNEF_FR
hnef_parse_piece_arg (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@in@*/
/*@notnull@*/
	struct hnef_type_piece	* const HNEF_RSTR tp,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD		* const	HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * tp, * fail_read@*/
{
	const char	* HNEF_RSTR arg2,
			* HNEF_RSTR arg3;

	assert	(NULL != wordv);
	assert	(NULL != tp);
	assert	(NULL != fail_read);

	if	(wordc < (size_t)3
/*@i2@*/\
	||	gleip_line_empty(wordv[2]))
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}
	arg2	= wordv[2];
	arg3	= NULL;

	if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPT_EDGE))
	{
		tp->capt_edge	= HNEF_TRUE;
		return		HNEF_FR_SUCCESS;
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPT_LOSS))
	{
		tp->capt_loss	= HNEF_TRUE;
		return		HNEF_FR_SUCCESS;
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_CUSTODIAL))
	{
		tp->custodial	= HNEF_TRUE;
		return		HNEF_FR_SUCCESS;
	}
	else if (0 == strcmp(arg2, HNEF_RREAD_L_DBL_TRAP))
	{
		tp->dbl_trap	= HNEF_TRUE;
		return		HNEF_FR_SUCCESS;
	}
	else if (0 == strcmp(arg2, HNEF_RREAD_L_DBL_TRAP_CAPT))
	{
		tp->dbl_trap_capt	= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
	else if (0 == strcmp(arg2, HNEF_RREAD_L_DBL_TRAP_COMPL))
	{
		tp->dbl_trap_compl	= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
	else if (0 == strcmp(arg2, HNEF_RREAD_L_DBL_TRAP_EDGE))
	{
		tp->dbl_trap_edge	= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
	else if (0 == strcmp(arg2, HNEF_RREAD_L_DBL_TRAP_ENCL))
	{
		tp->dbl_trap_encl	= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
/*@i1@*/\
	else if (0 == strcmp(arg2, HNEF_RREAD_L_ESCAPE))
	{
		tp->escape		= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
/*
 * NOTE:	Additional 3-word arguments go here.
 *		"piece <tp->bit> <arg2>"
 */

	if	(wordc > (size_t)3
/*@i2@*/\
	||	!gleip_line_empty(wordv[3]))
	{
		arg3	= wordv[3];
	}
	else
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}

	if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPTURES))
	{
		return	hnef_parse_type_bitu8(arg3, & tp->captures,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_NORETURN))
	{
		return	hnef_parse_type_bitu8(arg3, & tp->noreturn,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_OCCUPIES))
	{
		return	hnef_parse_type_bitu8(arg3, & tp->occupies,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPT_SIDES))
	{
		return	hnef_parse_type_ushrt(arg3, & tp->capt_sides,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_OWNER))
	{
		return	hnef_parse_type_ushrt(arg3, & tp->owner,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_UI_BIT))
	{
		const enum HNEF_FR fr	= hnef_parse_type_bitu8
					(arg3, & tp->ui_bit, fail_read);
		if	(!hnef_fr_good(fr))
		{
			return	fr;
		}
		else if	(HNEF_BIT_U8_EMPTY != tp->ui_bit
		&&	!hnef_single_bit((unsigned int)tp->ui_bit))
		{
			* fail_read	= HNEF_RREAD_FAIL_BIT_SINGLE;
			return		HNEF_FR_SUCCESS;
		}
		else
		{
			return	HNEF_FR_SUCCESS;
		}
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_DBL_TRAP_SQ))
	{
		return	hnef_parse_type_bitu8(arg3,
			& tp->dbl_trap_squares, fail_read);
	}
/*
 * NOTE:	Additional 4-word arguments go here.
 *		"piece <tp->bit> <arg2> <arg3>"
 */
	else
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_UNK;
		return		HNEF_FR_SUCCESS;
	}
}

/*
 * Parses "square x <arg>", where x is a single bit, and ts is the
 * square corresponding to bit x.
 */
static
enum HNEF_FR
hnef_parse_square_arg (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@in@*/
/*@notnull@*/
	struct hnef_type_square	* const HNEF_RSTR ts,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD		* const	HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * ts, * fail_read@*/
{
	const char	* arg2,
			* arg3;

	assert	(NULL != wordv);
	assert	(NULL != ts);
	assert	(NULL != fail_read);

	if	(wordc < (size_t)3
/*@i2@*/\
	||	gleip_line_empty(wordv[2]))
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}
	arg2	= wordv[2];
	arg3	= NULL;

	if	(0 == strcmp(arg2, HNEF_RREAD_L_ESCAPE))
	{
		ts->escape		= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
/*
 * NOTE:	Additional 3-word arguments go here.
 *		"square <ts->bit> <arg2>"
 */

	if	(wordc > (size_t)3
/*@i2@*/\
	||	!gleip_line_empty(wordv[3]))
	{
		arg3	= wordv[3];
	}
	else
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}

	if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPTURES))
	{
		return	hnef_parse_type_bitu8(arg3, & ts->captures,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPT_SIDES))
	{
		return	hnef_parse_type_ushrt(arg3, & ts->capt_sides,
			fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_CAPT_SIDES_PIECES))
	{
		return	hnef_parse_type_bitu8(arg3,
			& ts->capt_sides_pieces, fail_read);
	}
	else if	(0 == strcmp(arg2, HNEF_RREAD_L_UI_BIT))
	{
		const enum HNEF_FR fr	= hnef_parse_type_bitu8
					(arg3, & ts->ui_bit, fail_read);
		if	(!hnef_fr_good(fr))
		{
			return	fr;
		}
		else if	(HNEF_BIT_U8_EMPTY != ts->ui_bit
		&&	!hnef_single_bit((unsigned int)ts->ui_bit))
		{
			* fail_read	= HNEF_RREAD_FAIL_BIT_SINGLE;
			return		HNEF_FR_SUCCESS;
		}
		else
		{
			return	HNEF_FR_SUCCESS;
		}
	}
/*
 * NOTE:	Additional 4-word arguments go here.
 *		"square <ts->bit> <arg2> <arg3>"
 */
	else
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_UNK;
		return		HNEF_FR_SUCCESS;
	}
}

/*
 * Parses "piece/square <arg>", where <arg> is either a HNEF_BIT_U8 or
 * an argument.
 *
 * read_sq is true if we're parsing a square bit; false if it's a piece
 * bit.
 */
static
enum HNEF_FR
hnef_parse_type (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@in@*/
/*@notnull@*/
	struct hnef_game	* const HNEF_RSTR game,
	const HNEF_BOOL		read_sq,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * game, * fail_read@*/
{
	const char	* arg1;

	assert	(NULL != wordv);
	assert	(NULL != game);
	assert	(NULL != fail_read);

	if	(wordc < (size_t)2
/*@i2@*/\
	||	gleip_line_empty(wordv[1]))
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
		return		HNEF_FR_SUCCESS;
	}
	arg1	= wordv[1];

	if	(0 == strcmp(arg1, HNEF_RREAD_L_BIT))
	{
		return	hnef_parse_type_bit(wordv, wordc, game, read_sq,
					fail_read);
	}
	else
	{
		HNEF_BIT_U8	bit;
		{
			unsigned short	num;
			if	(!hnef_texttoushort(arg1, & num))
			{
				* fail_read = HNEF_RREAD_FAIL_STRTOL;
				return	HNEF_FR_SUCCESS;
			}
/*@i1@*/\
			else if	(num > (unsigned short)128
			||	!hnef_single_bit((unsigned int)num))
			{
				* fail_read =
					HNEF_RREAD_FAIL_BIT_SINGLE;
				return	HNEF_FR_SUCCESS;
			}
			bit	= (HNEF_BIT_U8)num;
		}
		if	(read_sq)
		{
			struct hnef_type_square * const ts =
				hnef_type_square_get(game->rules, bit);
			if	(HNEF_BIT_U8_EMPTY == ts->bit)
			{
				* fail_read =
					HNEF_RREAD_FAIL_TYPE_UNDEF;
				return	HNEF_FR_SUCCESS;
			}
			return hnef_parse_square_arg	(wordv, wordc,
						ts, fail_read);
		}
		else
		{
			struct hnef_type_piece * const tp =
				hnef_type_piece_get(game->rules, bit);
			if	(HNEF_BIT_U8_EMPTY == tp->bit)
			{
				* fail_read =
					HNEF_RREAD_FAIL_TYPE_UNDEF;
				return	HNEF_FR_SUCCESS;
			}
			return hnef_parse_piece_arg	(wordv, wordc,
						tp, fail_read);
		}
	}
}

static
enum HNEF_FR
hnef_parse_strarg (
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@out@*/
/*@notnull@*/
	char		* HNEF_RSTR * const HNEF_RSTR ptr,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD	* const HNEF_RSTR fail_read
	)
/*@modifies ptr, * fail_read@*/
{
	const char *	arg1;

	assert	(NULL != wordv);
	assert	(NULL != ptr);
	assert	(NULL != fail_read);

	if	(wordc < (size_t)2
/*@i2@*/\
	||	gleip_line_empty(wordv[1]))
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_MISS;
/* splint bug: pointer to pointer */ /*@i1@*/\
		return		HNEF_FR_SUCCESS;
	}
	arg1	= wordv[1];

	* ptr	= gleip_line_alloc_cpy(arg1);
	return	NULL == * ptr
		? HNEF_FR_FAIL_ALLOC
		: HNEF_FR_SUCCESS;
}

/*
 * Parses a line in the ruleset file.
 */
static
enum HNEF_FR
hnef_parseline (
/*@in@*/
/*@notnull@*/
	struct hnef_game		* const HNEF_RSTR game,
/*@in@*/
/*@notnull@*/
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t	wordc,
/*@in@*/
/*@notnull@*/
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
/*@globals errno@*/
/*@modifies errno, * game, * fail_read@*/
{
	/*
	 * g is NULL until bwidth and bheight in info are set.
	 */
	const char		* HNEF_RSTR arg0;
	struct hnef_ruleset	* HNEF_RSTR rules	= NULL;

	assert	(NULL != game);
	assert	(NULL != wordv);
	assert	(NULL != fail_read);
	assert	(wordc > 0);
/*@i2@*/\
	assert	(!gleip_line_empty(wordv[0]));

	rules	= game->rules;

	arg0	= wordv[0];

	if	(0 == strcmp(arg0, HNEF_RREAD_L_ID))
	{
		if	(NULL != rules->id)
		{
			* fail_read	= HNEF_RREAD_FAIL_ARG_DUP;
			return		HNEF_FR_SUCCESS;
		}
		return	hnef_parse_strarg(wordv, wordc,
					& rules->id, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_NAME))
	{
		if	(NULL != rules->name)
		{
			* fail_read	= HNEF_RREAD_FAIL_ARG_DUP;
			return		HNEF_FR_SUCCESS;
		}
		return	hnef_parse_strarg(wordv, wordc,
					& rules->name, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_UITHEME))
	{
		if	(NULL != rules->ui_theme)
		{
			* fail_read	= HNEF_RREAD_FAIL_ARG_DUP;
			return		HNEF_FR_SUCCESS;
		}
		return	hnef_parse_strarg(wordv, wordc,
					& rules->ui_theme, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_WIDTH))
	{
		return	hnef_parse_size	(game->rules, wordv, wordc,
					& rules->bwidth, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_HEIGHT))
	{
		return	hnef_parse_size	(game->rules, wordv, wordc,
					& rules->bheight, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_PIECES))
	{
		return	hnef_parse_arrb	(wordv, wordc, game,
					HNEF_FALSE, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_SQUARES))
	{
		return	hnef_parse_arrb	(wordv, wordc, game,
					HNEF_TRUE, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_PIECE))
	{
		return	hnef_parse_type	(wordv, wordc, game,
					HNEF_FALSE, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_SQUARE))
	{
		return	hnef_parse_type	(wordv, wordc, game,
					HNEF_TRUE, fail_read);
	}
	else if	(0 == strcmp(arg0, HNEF_RREAD_L_FREPEAT))
	{
		game->rules->forbid_repeat	= HNEF_TRUE;
		return			HNEF_FR_SUCCESS;
	}
/*
 * NOTE:	Additional arguments go here.
 *		"<arg0>"
 */
	else
	{
		* fail_read	= HNEF_RREAD_FAIL_ARG_UNK;
		return		HNEF_FR_SUCCESS;
	}
}

/*
 * Parses a tokenized line into a game struct.
 *
 * This function only parses a single line. The ruleset presumably
 * consists of several lines, so you will have to call this function
 * once per line until you have exhausted all lines.
 *
 * You have to tokenize the line yourself before calling. `wordc` is
 * the length of the `wordv` array (just like `argc` / `argv` in
 * `main()`, except the first token has index 0 here). If you attempt to
 * pass an untokenized whole line as `wordv` with `wordc == 1`, then
 * this function will consider the whole line to be the first (and only)
 * token and most likely fail.
 *
 * Usage:
 *
 *	while	(...has_more_tokens...)
 *	{
 *		hnef_game_parse_line_tokens(...tokens...);
 *	}
 *
 * A `fail_read` failure indicates a non-fatal failure to parse a
 * ruleset line (the ruleset file is malformed, or the line has not been
 * properly tokenized).
 *
 * A `HNEF_FR` failure is supposed to be fatal, but the game object can
 * be reused after this function returns with a `HNEF_FR` failure by
 * calling `hnef_game_clear()`.
 */
enum HNEF_FR
hnef_game_parse_line_tokens (
	struct hnef_game	* const HNEF_RSTR game,
	const char	* const HNEF_RSTR * const HNEF_RSTR wordv,
	const size_t		wordc,
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
{
	assert	(NULL != game);
	assert	(NULL != wordv);
	assert	(NULL != fail_read);

	* fail_read	= HNEF_RREAD_SUCCESS;

	game->opt_valid_known	= HNEF_FALSE;
	game->opt_valid_code	= HNEF_RVALID_SUCCESS;

	if	(wordc < (size_t)1)
	{
		return	HNEF_FR_SUCCESS;
	}

	return	hnef_parseline(game, wordv, wordc, fail_read);
}

/*
 * Delegates `hnef_game_parse_tokens()` with a `gleip_lines` struct,
 * which contains the tokenized line.
 */
enum HNEF_FR
hnef_game_read_line_single (
	struct hnef_game		* const HNEF_RSTR game,
	const struct gleip_lines	* const GLEIP_RSTR line,
	enum HNEF_RREAD			* const HNEF_RSTR fail_read
	)
{
	assert	(NULL != game);
	assert	(NULL != line);
	assert	(NULL != fail_read);

	/*
	 * This reset `opt_valid_known`.
	 */
	return	hnef_game_parse_line_tokens(game,
	(const char * const GLEIP_RSTR * const GLEIP_RSTR)line->arr,
				gleip_lines_size(line), fail_read);
}

/*
 * Parses a ruleset file into a game struct.
 *
 * Note that this function does not guarantee that the ruleset is valid.
 * A successful return only indicates the ruleset file was correctly
 * formatted.
 *
 * Usage:
 *
 *	game = ...
 *	file = ... // Ruleset file
 *	hnef_game_clear		(game);
 *	hnef_game_read_file	(game, file...);	// Check errors
 *	if	(hnef_rvalid_good(hnef_game_valid(game)))
 *	{
 *		hnef_game_init_done	(game);		// Check errors
 *	}
 *	else
 *	{
 *		hnef_game_clear		(game);
 *	}
 *
 * See `hnef_game_read_line_single()` for details.
 *
 * -	`hnef_fr_fail_io()` should be treated as non-fatal errors
 *	(failure to read the file).
 * -	Other failures (`HNEF_FR_FAIL_ALLOC`) should be treated as
 *	fatal.
 *
 * Note that this only performs one step in initializing a game struct.
 * You may to use `hnef_game_init_file()` instead, which does the full
 * initialization.
 */
enum HNEF_FR
hnef_game_read_file (
	struct hnef_game	* const HNEF_RSTR game,
	FILE			* const HNEF_RSTR file,
	struct gleip_lines	* const GLEIP_RSTR line,
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
{
	enum HNEF_FR	fr	= HNEF_FR_SUCCESS;
	int		retval	= 1;

	assert	(NULL != game);
	assert	(NULL != file);
	assert	(NULL != line);
	assert	(NULL != fail_read);

	* fail_read	= HNEF_RREAD_SUCCESS;

	game->opt_valid_known	= HNEF_FALSE;
	game->opt_valid_code	= HNEF_RVALID_SUCCESS;

	while	(EOF != retval
/*@i1@*/\
	&&	!feof(file))
	{
		gleip_lines_rm(line);
		retval	= gleip_lineread_interp_file_def(line, file, 0);
/*@i1@*/\
		if	(!retval)
		{
/*@i1@*/\
			return	ferror(file)
				? HNEF_FR_FAIL_IO_FILE_READ
				: HNEF_FR_FAIL_ALLOC;
		}

		fr	= hnef_game_read_line_single
			(game, line, fail_read);
		if	(!hnef_fr_good(fr)
		||	!hnef_rread_good(* fail_read))
		{
			return	fr;
		}
	}
	return	HNEF_FR_SUCCESS;
}

/*
 * Works like `hnef_game_read_file()` except it reads a C-string
 * (`content`), which has been read from a file or similar. Thus you
 * don't have to operate on a `FILE *`.
 *
 * `content` should consist of lines that are delimited by '\n'.
 */
enum HNEF_FR
hnef_game_read_line (
	struct hnef_game	* const HNEF_RSTR game,
	const char		* const content,
	struct gleip_lines	* const GLEIP_RSTR line,
	enum HNEF_RREAD		* const HNEF_RSTR fail_read
	)
{
	const char	* ptr	= content;
	enum HNEF_FR	fr	= HNEF_FR_SUCCESS;
	int		retval	= 1;

	assert	(NULL != game);
	assert	(NULL != content);
	assert	(NULL != line);
	assert	(NULL != fail_read);

	* fail_read	= HNEF_RREAD_SUCCESS;

	game->opt_valid_known	= HNEF_FALSE;
	game->opt_valid_code	= HNEF_RVALID_SUCCESS;

	while	(EOF != retval)
	{
		gleip_lines_rm(line);
/*
 * Here SPLint thinks we are modifying `* content`, but this is because
 * SPLint doesn't understand pointer to pointer. In
 * `hnef_line_interp_line()`, we modify the value pointed to by a
 * pointer to a C-string (`const char * *`, that is: we do pointer
 * arithmetic on the C-string pointer when iterating over the chars) --
 * we don't actually modify the chars in the C-string (which may be a
 * read-only memory location).
 */
/*@i1@*/\
		retval = gleip_lineread_interp_cstr_def(line, & ptr, 0);
/*@i1@*/\
		if	(!retval)
		{
			return	HNEF_FR_FAIL_ALLOC;
		}

		fr	= hnef_game_read_line_single
			(game, line, fail_read);
		if	(!hnef_fr_good(fr)
		||	!hnef_rread_good(* fail_read))
		{
			return	fr;
		}
	}
	return	HNEF_FR_SUCCESS;
}

/*
 * Convenience function for reading and fully initializing a game from a
 * file.
 *
 * Clears the game, reads it, and if valid, finished the initialization.
 * If anything goes wrong, the game is cleared, and a failure is
 * indicated by `HNEF_FR`, `HNEF_RREAD` or `HNEF_RVALID`.
 *
 * -	If the `HNEF_FR` failure is `hnef_fr_fail_io()`, then it's
 *	non-fatal (probably failed to read from the file). Else it
 *	should be considered fatal (probably out of memory).
 * -	If `HNEF_RREAD` indicates a failure, `line` indicates the line
 *	that the program was parsing when the failure occured. This may
 *	provide valuable information to the user.
 *
 * The game is valid if the returned `HNEF_FR`, `fail_read` and
 * `fail_valid` all indicate success. This implies that the game is
 * valid and fully initialized (if they all return success, then you can
 * `assert()` that the game is valid). If any of them does not indicate
 * success, then the game is invalid and cleared.
 */
enum HNEF_FR
hnef_game_init_file (
	struct hnef_game	* const HNEF_RSTR game,
	FILE			* const HNEF_RSTR file,
	struct gleip_lines	* const GLEIP_RSTR line,
	enum HNEF_RREAD		* const HNEF_RSTR fail_read,
	enum HNEF_RVALID	* const HNEF_RSTR fail_valid
	)
{
	enum HNEF_FR	fr	= HNEF_FR_SUCCESS;

	assert	(NULL != game);
	assert	(NULL != file);
	assert	(NULL != line);
	assert	(NULL != fail_read);
	assert	(NULL != fail_valid);

	* fail_read	= HNEF_RREAD_SUCCESS;
	* fail_valid	= HNEF_RVALID_SUCCESS;

	/*
	 * Sets `opt_valid_known` to false.
	 */
	hnef_game_clear	(game);

	fr	= hnef_game_read_file(game, file, line, fail_read);
	if	(!hnef_fr_good(fr)
	||	!hnef_rread_good(* fail_read))
	{
		hnef_game_clear(game);
		return	fr;
	}

	if	(hnef_rvalid_good(* fail_valid = hnef_game_valid(game)))
	{
		fr	= hnef_game_init_done(game);
		if	(!hnef_fr_good(fr))
		{
			hnef_game_clear(game);
		}
		return	fr;
	}
	else
	{
		hnef_game_clear(game);
		return	fr;
	}
}

/*
 * Exactly like `hnef_game_init_file()` but uses a C-string (`content`).
 *
 * -	Unlike `hnef_game_init_file()`, all `HNEF_FR` failures are
 *	fatal here, because they can never indicate an IO failure.
 */
enum HNEF_FR
hnef_game_init_line (
	struct hnef_game	* const HNEF_RSTR game,
	const char		* const content,
	struct gleip_lines	* const GLEIP_RSTR line,
	enum HNEF_RREAD		* const HNEF_RSTR fail_read,
	enum HNEF_RVALID	* const HNEF_RSTR fail_valid
	)
{
	enum HNEF_FR	fr	= HNEF_FR_SUCCESS;

	assert	(NULL != game);
	assert	(NULL != content);
	assert	(NULL != line);
	assert	(NULL != fail_read);
	assert	(NULL != fail_valid);

	* fail_read	= HNEF_RREAD_SUCCESS;
	* fail_valid	= HNEF_RVALID_SUCCESS;

	/*
	 * Sets `opt_valid_known` to false.
	 */
	hnef_game_clear	(game);

	fr	= hnef_game_read_line(game, content, line, fail_read);
	if	(!hnef_fr_good(fr)
	||	!hnef_rread_good(* fail_read))
	{
		hnef_game_clear(game);
		return	fr;
	}

	if	(hnef_rvalid_good(* fail_valid = hnef_game_valid(game)))
	{
		fr	= hnef_game_init_done(game);
		if	(!hnef_fr_good(fr))
		{
			hnef_game_clear(game);
		}
		return	fr;
	}
	else
	{
		hnef_game_clear(game);
		return	fr;
	}
}

